import typing

from .bytes_utils import join, split


def encrypt_data(
    in_data, in_encrypt: typing.Callable[[bytes, bytes], bytes]
) -> typing.Tuple[str, bytes]:
    if list(in_data.keys()) == ["str"]:
        return in_data["str"], bytes()
    tmp = encrypt_data(in_data["rest"], in_encrypt)
    rest = join(tmp[0], tmp[1])
    encrypted = in_encrypt(rest, in_data["pass"].encode())
    return in_data["str"], encrypted


def decrypt_data(
    in_rest: bytes,
    in_pass: str,
    in_decrypt: typing.Callable[[bytes, bytes], bytes | None],
) -> None | typing.Tuple[str, bytes]:
    res = in_decrypt(in_rest, in_pass.encode())
    if res is None:
        return None
    return split(res)



import typing
import secrets

from .common import (
    derive_key,
    xor_bytes,
    hash_bytes,
    digest_size,
    merge_data_and_signature,
    split_data_and_signature,
)


def get_encrypt(
    proc_spices: typing.List[bytes],
    signature_spices: typing.List[bytes],
    scrypt_params,
    signature_params,
) -> typing.Callable[[bytes, bytes], bytes]:
    assert proc_spices  # nosec B101
    assert signature_spices  # nosec B101

    def _encrypt(in_bytes: bytes, in_pass: bytes) -> bytes:
        signature_spice = secrets.choice(signature_spices)
        signature = hash_bytes(in_bytes + signature_spice, signature_params)
        merged = merge_data_and_signature(in_bytes, signature)
        proc_spice = secrets.choice(proc_spices)
        key = derive_key(
            password=in_pass + proc_spice, dklen=len(merged), **scrypt_params
        )
        return xor_bytes(merged, key)

    return _encrypt


def get_decrypt(
    proc_spices: typing.List[bytes],
    signature_spices: typing.List[bytes],
    scrypt_params,
    signature_params,
) -> typing.Callable[[bytes, bytes], bytes | None]:
    assert proc_spices  # nosec B101
    assert signature_spices  # nosec B101

    def _decrypt(in_bytes: bytes, in_pass: bytes) -> bytes | None:
        for proc_spice in proc_spices:
            key = derive_key(
                password=in_pass + proc_spice, dklen=len(in_bytes), **scrypt_params
            )
            data = xor_bytes(in_bytes, key)
            decrypted, signature = split_data_and_signature(
                data, digest_size(signature_params)
            )

            if any(
                hash_bytes(decrypted + _, signature_params) == signature
                for _ in signature_spices
            ):
                return decrypted
        return None

    return _decrypt



import hashlib
import typing


def digest_size(params) -> int:
    hasher = hashlib.new(**params["hasher"])
    res = hasher.digest_size
    if res > 0:
        return res
    return params["digest"]["length"]


def hash_bytes(in_bytes: bytes, params) -> bytes:
    hasher = hashlib.new(**params["hasher"])
    hasher.update(in_bytes)
    res = hasher.digest(**params["digest"])
    return res


def derive_key(**kwargs):
    return hashlib.scrypt(**kwargs)


def xor_bytes(in_bytes: bytes, in_key: bytes) -> bytes:
    """xors the in_bytes with a sequence of bytes with in_key"""
    return bytes(_d ^ _k for (_d, _k) in zip(in_bytes, in_key, strict=True))


def merge_data_and_signature(in_data: bytes, in_signature: bytes) -> bytes:
    return in_data + in_signature


def split_data_and_signature(
    in_bytes: bytes, signature_size: int
) -> typing.Tuple[bytes, bytes]:
    assert len(in_bytes) >= signature_size  # nosec B101
    data = in_bytes[:-signature_size]
    signature = in_bytes[-signature_size:]
    return data, signature



def check_kwargs(allowed_params, **kwargs):
    for cur_param in kwargs:
        if cur_param not in allowed_params:
            raise TypeError(
                f"{cur_param} is an invalid keyword argument. "
                f"Valid arguments are {allowed_params}."
            )



import secrets

from ...encryption_algorithms.simple import common
from ... import bytestr_utils as bsu
from ... import bytes_utils as bu
from ... import puzzle_data_encryption as pde
from ... import run_puzzle as rp

MODULES = ["hashlib", "base64", "sys", "typing"]

OBJECTS = [
    common.hash_bytes,
    common.derive_key,
    common.split_data_and_signature,
    common.digest_size,
    common.xor_bytes,
    bsu.bytestr_to_bytes,
    bu.bytes_to_int,
    bu.split,
    pde.decrypt_data,
    rp.run_puzzle,
]


def scrypt_params(**kwargs):
    in_params = kwargs.get("scrypt_params", {})
    default = {"salt": secrets.token_bytes(16), "n": 2**14, "r": 8, "p": 1}
    if "maxmem" in in_params:
        default["maxmem"] = in_params["maxmem"]
    res = {_k: in_params.get(_k, _v) for _k, _v in default.items()}
    if "maxmem" not in res:
        res["maxmem"] = (128 + 100) * res["n"] * res["r"] * res["p"]
    return res


def signature_params(**kwargs):
    res = kwargs.get("signature_params", {"hasher": {"name": "sha512"}, "digest": {}})
    if "digest" not in res:
        res["digest"] = {}
    return res


def scrypt_params_to_code_str(**kwargs) -> str:
    pieces = [f'"{_k}":{_v}' for _k, _v in kwargs.items() if _k != "salt"]
    salt_str = '"' + bsu.bytes_to_bytestr(kwargs["salt"]) + '"'
    pieces.append(f'"salt":bytestr_to_bytes({salt_str})')
    return f"_SCRYPT_PARAMS = {{{','.join(pieces)}}}"



import secrets
import typing

from ... import bytestr_utils as bu
from ...encryption_algorithms.simple import spiced as sse
from . import common as csc
from ..check_kwargs import check_kwargs


def _get_some_spices():
    return [secrets.token_bytes(3) for _ in range(20)]


def _list_of_bytes_to_str(in_list: typing.List[bytes]) -> str:
    return str([bu.bytes_to_bytestr(_) for _ in in_list])


def _list_of_bytes_to_codestr(in_list: typing.List[bytes]) -> str:
    raw: str = _list_of_bytes_to_str(in_list)
    return f"[bytestr_to_bytes(_) for _ in {raw}]"


class Spiced:
    def __init__(self, **kwargs):
        check_kwargs(
            {"scrypt_params", "signature_params", "proc_spices", "signature_spices"},
            **kwargs,
        )
        self._scrypt_params = csc.scrypt_params(**kwargs)
        self._signature_params = csc.signature_params(**kwargs)
        self._proc_spices = kwargs.get("proc_spices", _get_some_spices())
        self._signature_spices = kwargs.get("signature_spices", _get_some_spices())

    def get_modules(self) -> typing.List[str]:
        return csc.MODULES

    def get_encrypt(self):
        return sse.get_encrypt(
            self._proc_spices,
            self._signature_spices,
            self._scrypt_params,
            self._signature_params,
        )

    def get_needed_objects(self):
        return csc.OBJECTS + [
            sse.get_decrypt,
        ]

    def get_constants_str(
        self,
    ) -> str:
        proc_spices: str = (
            f"_PROC_SPICES = {_list_of_bytes_to_codestr(self._proc_spices)}"
        )
        signature_spices: str = (
            f"_SIGNATURE_SPICES = {_list_of_bytes_to_codestr(self._signature_spices)}"
        )
        _scrypt_params = csc.scrypt_params_to_code_str(**self._scrypt_params)
        _signature_params = "_SIGNATURE_PARAMS = " + repr(self._signature_params)
        decrypt: str = (
            "_DECRYPT = get_decrypt("
            "_PROC_SPICES, "
            "_SIGNATURE_SPICES, "
            "_SCRYPT_PARAMS, _SIGNATURE_PARAMS)"
        )
        return "\n".join(
            [proc_spices, signature_spices, _scrypt_params, _signature_params, decrypt]
        )
